@@ -1,5 +1,4 @@
 import copy
-import io
 import itertools
 import json
 import logging
@@ -12,6 +11,8 @@ import uuid
 
 from datamart_materialize.d3m import d3m_metadata
 from datamart_materialize import types
+from datamart_profiler.temporal import get_temporal_resolution, \
+    temporal_aggregation_keys
 
 
 logger = logging.getLogger(__name__)
@@ -30,23 +31,38 @@ class _UniqueIndexKey(object):
 UNIQUE_INDEX_KEY = _UniqueIndexKey()
 
 
-temporal_resolutions = [
-    'second',
-    'minute',
-    'hour',
-    'date'
-]
-
-
-temporal_resolution_format = {
-    'second': '%Y-%m-%d %H:%M:%S',
-    'minute': '%Y-%m-%d %H:%M',
-    'hour': '%Y-%m-%d %H',
-    'date': '%Y-%m-%d'
+temporal_resolutions_priorities = {
+    n: i
+    for i, n in enumerate(reversed(list(temporal_aggregation_keys)))
 }
 
 
-def convert_data_types(data, columns, columns_metadata, drop=False):
+def _transform_index(level, func):
+    def inner(index):
+        if isinstance(index, pd.MultiIndex):
+            return index.set_levels(
+                [index.levels[i] if i != level
+                 else func(index.levels[i])
+                 for i in range(len(index.levels))]
+            )
+        else:
+            return func(index)
+
+    return inner
+
+
+def _transform_data_index(data, level, func):
+    if isinstance(data.index, pd.MultiIndex):
+        data.index = data.index.set_levels(
+            [data.index.levels[i] if i != level
+             else func(data.index.levels[i])
+             for i in range(len(data.index.levels))]
+        )
+    else:
+        data.index = func(data.index)
+
+
+def set_data_index(data, columns, columns_metadata, drop=False):
     """
     Converts columns in a dataset (pandas.DataFrame) to their corresponding
     data types, based on the provided metadata.
@@ -58,40 +74,33 @@ def convert_data_types(data, columns, columns_metadata, drop=False):
         inplace=True
     )
 
-    for i, index in enumerate(columns):
-        column = columns_metadata[index]
+    for i, col_idx in enumerate(columns):
+        column = columns_metadata[col_idx]
         if types.DATE_TIME in column['semantic_types']:
-            if isinstance(data.index, pd.MultiIndex):
-                data.index = data.index.set_levels(
-                    [data.index.levels[j] if j != i
-                     else pd.to_datetime(data.index.levels[j], errors='coerce')
-                     for j in range(len(data.index.levels))]
-                )
-            else:
-                data.index = pd.to_datetime(data.index, errors='coerce')
+            _transform_data_index(
+                data, i,
+                lambda idx: pd.to_datetime(idx.map(str), errors='coerce'),
+            )
         elif column['structural_type'] == types.INTEGER:
-            if isinstance(data.index, pd.MultiIndex):
-                data.index = data.index.set_levels(
-                    [data.index.levels[j] if j != i
-                     else pd.to_numeric(data.index.levels[j], errors='coerce', downcast='integer')
-                     for j in range(len(data.index.levels))]
-                )
-            else:
-                data.index = pd.to_numeric(data.index, errors='coerce', downcast='integer')
+            _transform_data_index(
+                data, i,
+                lambda idx: pd.to_numeric(idx, errors='coerce', downcast='integer'),
+            )
         elif column['structural_type'] == types.FLOAT:
-            if isinstance(data.index, pd.MultiIndex):
-                data.index = data.index.set_levels(
-                    [data.index.levels[j] if j != i
-                     else pd.to_numeric(data.index.levels[j], errors='coerce', downcast='float')
-                     for j in range(len(data.index.levels))]
-                )
-            else:
-                data.index = pd.to_numeric(data.index, errors='coerce', downcast='float')
+            _transform_data_index(
+                data, i,
+                lambda idx: pd.to_numeric(idx, errors='coerce', downcast='float'),
+            )
+        elif column['structural_type'] == types.TEXT:
+            _transform_data_index(
+                data, i,
+                lambda idx: idx.str.lower(),
+            )
 
     return data
 
 
-def match_temporal_resolutions(input_data, companion_data):
+def match_temporal_resolutions(input_data, companion_data, temporal_resolution=None):
     """Matches the resolutions between the datasets.
 
     This takes in example indexes, and returns a function to update future
@@ -100,46 +109,109 @@ def match_temporal_resolutions(input_data, companion_data):
     """
 
     if isinstance(input_data.index, pd.MultiIndex):
-        # TODO: support MultiIndex
-        pass
+        # Find which levels are temporal
+        funcs = []
+        for i, lvl in enumerate(input_data.index.levels):
+            if isinstance(lvl, pd.DatetimeIndex):
+                funcs.append(
+                    match_column_temporal_resolutions(
+                        input_data.index,
+                        companion_data.index,
+                        i,
+                        temporal_resolution,
+                    )
+                )
+            else:
+                funcs.append(lambda x: x)
+
+        return lambda idx: idx.set_levels(
+            func(lvl)
+            for lvl, func in zip(idx.levels, funcs)
+        )
     elif (isinstance(input_data.index, pd.DatetimeIndex)
           and isinstance(companion_data.index, pd.DatetimeIndex)):
-        return match_column_temporal_resolutions(input_data.index, companion_data.index)
+        return match_column_temporal_resolutions(
+            input_data.index,
+            companion_data.index,
+            0,
+            temporal_resolution,
+        )
 
-    return lambda input_idx, comp_idx: (input_idx, comp_idx)  # no-op
+    return lambda idx: idx  # no-op
 
 
-def match_column_temporal_resolutions(index_1, index_2):
+def match_column_temporal_resolutions(index_1, index_2, level,
+                                      temporal_resolution=None):
     """Matches the resolutions between the dataset indices.
     """
 
-    resolution_1 = check_temporal_resolution(index_1)
-    resolution_2 = check_temporal_resolution(index_2)
-    if (temporal_resolutions.index(resolution_1) >
-            temporal_resolutions.index(resolution_2)):
-        # Change resolution of second index to the first's
-        fmt = temporal_resolution_format[resolution_1]
-        return lambda idx1, idx2: (idx1, idx2.strftime(fmt))
+    if not (index_1.is_all_dates and index_2.is_all_dates):
+        return lambda idx: idx
+
+    # Keep in sync with search.get_joinable_datasets()
+
+    # Use the provided resolution
+    if temporal_resolution is not None:
+        key = temporal_aggregation_keys[temporal_resolution]
+        logger.info("Temporal alignment: requested '%s'", temporal_resolution)
+        if isinstance(key, str):
+            return _transform_index(level, lambda idx: idx.strftime(key))
+        else:
+            return _transform_index(level, lambda idx: idx.map(key))
     else:
-        # Change resolution of first index to the second's
-        fmt = temporal_resolution_format[resolution_2]
-        _idx1 = index_1.strftime(fmt)  # Cache it for speed
-        return lambda idx1, idx2: (_idx1, idx2)
+        # Pick the more coarse of the two resolutions
+        resolution_1 = get_temporal_resolution(index_1[~index_1.isna()])
+        resolution_2 = get_temporal_resolution(index_2[~index_2.isna()])
+
+        if (temporal_resolutions_priorities[resolution_1] >
+                temporal_resolutions_priorities[resolution_2]):
+            # Change resolution of second index to the first's
+            logger.info("Temporal alignment: right to '%s'", resolution_1)
+            key = temporal_aggregation_keys[resolution_1]
+            if isinstance(key, str):
+                return _transform_index(level, lambda idx: idx.strftime(key))
+            else:
+                return _transform_index(level, lambda idx: idx.map(key))
+        else:
+            # Change resolution of first index to the second's
+            logger.info("Temporal alignment: left to '%s'", resolution_2)
+            key = temporal_aggregation_keys[resolution_2]
+            if isinstance(key, str):
+                return _transform_index(level, lambda idx: idx.strftime(key))
+            else:
+                return _transform_index(level, lambda idx: idx.map(key))
 
 
-def check_temporal_resolution(data):
-    """Returns the resolution of the temporal attribute.
+def _first(series):
+    return series.iloc[0]
+
+
+def _sum(series):
+    """Variant of numpy.sum() that returns nan for all-nan array.
+
+    That way it works similarly to numpy.mean(), numpy.max(), etc instead of
+    returning 0.
     """
-
-    if not data.is_all_dates:
-        return None
-    for res in temporal_resolutions[:-1]:
-        if len(set([getattr(x, res) for x in data[data.notnull()]])) > 1:
-            return res
-    return 'date'
+    if np.any(~np.isnan(series)):
+        return np.sum(series)
+    else:
+        return np.nan
 
 
-def perform_aggregations(data, original_columns):
+AGGREGATION_FUNCTIONS = {
+    'first': pd.NamedAgg('first', _first),
+    'mean': pd.NamedAgg('mean', np.mean),
+    'sum': pd.NamedAgg('sum', _sum),
+    'max': pd.NamedAgg('max', np.max),
+    'min': pd.NamedAgg('min', np.min),
+    'count': pd.NamedAgg('count', lambda s: (~s.isna()).sum()),
+}
+
+
+def perform_aggregations(
+    data, original_columns,
+    agg_functions=None, augment_columns_name=None,
+):
     """Performs group by on dataset after join, to keep the shape of the
     new, augmented dataset the same as the original, input data.
     """
@@ -148,27 +220,48 @@ def perform_aggregations(data, original_columns):
         col: idx for idx, col in enumerate(data.columns)
     }
 
-    def first(series):
-        return series.iloc[0]
-
     start = time.perf_counter()
     original_columns_set = set(original_columns)
+
+    provided_agg_functions = agg_functions
+    if provided_agg_functions:
+        provided_agg_functions = {
+            # Columns might have been renamed if conflicting, deal with that
+            augment_columns_name[col]:
+                # Turn single value into list
+                [funcs] if isinstance(funcs, str) else funcs
+            for col, funcs in provided_agg_functions.items()
+        }
+
     agg_functions = dict()
     for column in data.columns:
         if column == UNIQUE_INDEX_KEY or column in original_columns_set:
             # Just pick the first value
             # (they are all the same, from a single row in the original data)
-            agg_functions[column] = [pd.NamedAgg('first', first)]
+            agg_functions[column] = ['first']
+        elif provided_agg_functions:
+            try:
+                funcs = provided_agg_functions[column]
+            except KeyError:
+                pass
+            else:
+                agg_functions[column] = (
+                    [funcs] if isinstance(funcs, str)
+                    else funcs
+                )
         else:
             if ('int' in str(data.dtypes[column]) or
                     'float' in str(data.dtypes[column])):
-                agg_functions[column] = [
-                    pd.NamedAgg('mean', np.mean), pd.NamedAgg('sum', np.sum),
-                    pd.NamedAgg('max', np.max), pd.NamedAgg('min', np.min),
-                ]
+                agg_functions[column] = ['mean', 'sum', 'max', 'min']
             else:
                 # Just pick the first value
-                agg_functions[column] = [pd.NamedAgg('first', first)]
+                agg_functions[column] = ['first']
+
+    # Resolve names into functions using AGGREGATION_FUNCTIONS map
+    agg_functions = {
+        col: [AGGREGATION_FUNCTIONS[name] for name in names]
+        for col, names in agg_functions.items()
+    }
 
     # Perform group-by
     data = data.groupby(by=[UNIQUE_INDEX_KEY]).agg(agg_functions)
@@ -185,7 +278,7 @@ def perform_aggregations(data, original_columns):
 
     # Rename columns
     data.columns = [
-        col[0] if col[1] == 'first'
+        col[0] if col[1] == 'first' and len(agg_functions[col[0]]) <= 1
         else ' '.join(col[::-1]).strip()
         for col in data.columns
     ]
@@ -197,10 +290,14 @@ def perform_aggregations(data, original_columns):
 CHUNK_SIZE_ROWS = 10000
 
 
-def join(original_data, augment_data_path, original_metadata, augment_metadata,
-         destination_csv,
-         left_columns, right_columns,
-         how='left', columns=None, return_only_datamart_data=False):
+def join(
+    original_data, augment_data_path, original_metadata, augment_metadata,
+    destination_csv,
+    left_columns, right_columns,
+    how='left', columns=None,
+    agg_functions=None, temporal_resolution=None,
+    return_only_datamart_data=False,
+):
     """
     Performs a join between original_data (pandas.DataFrame)
     and augment_data (pandas.DataFrame) using left_columns and right_columns.
@@ -220,7 +317,7 @@ def join(original_data, augment_data_path, original_metadata, augment_metadata,
         original_join_columns_idx.append(left[0])
         augment_join_columns_idx.append(right[0])
 
-    original_data = convert_data_types(
+    original_data = set_data_index(
         original_data,
         original_join_columns_idx,
         original_metadata['columns'],
@@ -232,22 +329,16 @@ def join(original_data, augment_data_path, original_metadata, augment_metadata,
 
     logger.info("Performing join...")
 
-    # join columns
-    augment_join_columns = list()
-    for left, right in zip(left_columns, right_columns):
-        left_name = original_data.columns[left[0]]
-        right_name = augment_data_columns[right[0]]
-        if right_name == left_name:
-            right_name += '_r'
-        augment_join_columns.append(right_name)
-
     # Stream the data in
     augment_data_chunks = pd.read_csv(
         augment_data_path,
         error_bad_lines=False,
         chunksize=CHUNK_SIZE_ROWS,
     )
-    first_augment_data = next(augment_data_chunks)
+    try:
+        first_augment_data = next(augment_data_chunks)
+    except StopIteration:
+        raise AugmentationError("Empty augmentation data")
 
     # Columns to drop
     drop_columns = None
@@ -264,8 +355,10 @@ def join(original_data, augment_data_path, original_metadata, augment_metadata,
             )
         )
 
-    # Guess temporal resolutions
-    update_idx = match_temporal_resolutions(original_data, first_augment_data)
+    # Defer temporal alignment until reading the first block from companion
+    # (and converting it to the right data types!)
+    update_idx = None
+    original_data_res = None
 
     # Streaming join
     start = time.perf_counter()
@@ -275,18 +368,26 @@ def join(original_data, augment_data_path, original_metadata, augment_metadata,
             itertools.chain([first_augment_data], augment_data_chunks)
     ):
         # Convert data types
-        augment_data = convert_data_types(
+        augment_data = set_data_index(
             augment_data,
             augment_join_columns_idx,
             augment_metadata['columns'],
             drop=True,  # Drop the join columns on that side (avoid duplicates)
         )
 
+        if update_idx is None:
+            # Guess temporal resolutions (on first chunk)
+            update_idx = match_temporal_resolutions(
+                original_data,
+                augment_data,
+                temporal_resolution,
+            )
+            original_data_res = original_data.set_index(
+                update_idx(original_data.index)
+            )
+
         # Match temporal resolutions
-        original_data_res, augment_data = update_idx(
-            original_data,
-            augment_data,
-        )
+        augment_data.index = update_idx(augment_data.index)
 
         # Filter columns
         if drop_columns:
@@ -307,21 +408,20 @@ def join(original_data, augment_data_path, original_metadata, augment_metadata,
     join_ = pd.concat(join_)
     logger.info("Join completed in %.4fs", time.perf_counter() - start)
 
+    intersection = set(original_data.columns).intersection(set(first_augment_data.columns))
+
     # qualities
-    qualities_list = list()
+    qualities_list = []
 
     if return_only_datamart_data:
         # drop unique index
         join_.drop([UNIQUE_INDEX_KEY], axis=1, inplace=True)
 
         # drop columns from original data
-        drop_columns = list()
-        intersection = set(original_data.columns).intersection(set(first_augment_data.columns))
-        if len(intersection) > 0:
-            drop_columns = list(intersection)
-        drop_columns += list(set(original_data.columns).difference(intersection))
+        drop_columns = list(intersection)
+        drop_columns.extend(set(original_data.columns).difference(intersection))
         join_.drop(drop_columns, axis=1, inplace=True)
-        if len(intersection) > 0:
+        if intersection:
             rename = dict()
             for column in intersection:
                 rename[column + '_r'] = column
@@ -331,10 +431,18 @@ def join(original_data, augment_data_path, original_metadata, augment_metadata,
         join_.dropna(axis=0, how='all', inplace=True)
 
     else:
+        # map column names for the augmentation data
+        augment_columns_map = {
+            name: name + '_r' if name in intersection else name
+            for name in first_augment_data.columns
+        }
+
         # aggregations
         join_ = perform_aggregations(
             join_,
             list(original_data.columns),
+            agg_functions,
+            augment_columns_map,
         )
 
         # drop unique index
@@ -368,12 +476,10 @@ def join(original_data, augment_data_path, original_metadata, augment_metadata,
             column['name'] + '_r'
         ]
         # agg names
-        all_names = ['sum ' + name for name in names]
-        all_names += ['mean ' + name for name in names]
-        all_names += ['max ' + name for name in names]
-        all_names += ['min ' + name for name in names]
-        all_names += ['first ' + name for name in names]
-        all_names += names
+        all_names = itertools.chain(names, (
+            agg + ' ' + name
+            for agg, name in itertools.product(AGGREGATION_FUNCTIONS, names)
+        ))
         for name in all_names:
             column_metadata = copy.deepcopy(column)
             column_metadata['name'] = name
@@ -492,7 +598,7 @@ def augment(data, newdata, metadata, task, columns=None, destination=None,
     """
     Augments original data based on the task.
 
-    :param data: the data to be augmented, as bytes.
+    :param data: the data to be augmented, as binary file object.
     :param newdata: the path to the CSV file to augment with.
     :param metadata: the metadata of the data to be augmented.
     :param task: the augmentation task.
@@ -522,7 +628,7 @@ def augment(data, newdata, metadata, task, columns=None, destination=None,
     start = time.perf_counter()
     if task['augmentation']['type'] == 'join':
         output_metadata = join(
-            pd.read_csv(io.BytesIO(data), error_bad_lines=False),
+            pd.read_csv(data, error_bad_lines=False),
             newdata,
             metadata,
             task['metadata'],
@@ -530,11 +636,13 @@ def augment(data, newdata, metadata, task, columns=None, destination=None,
             task['augmentation']['left_columns'],
             task['augmentation']['right_columns'],
             columns=columns,
+            agg_functions=task['augmentation'].get('agg_functions'),
+            temporal_resolution=task['augmentation'].get('temporal_resolution'),  # look
             return_only_datamart_data=return_only_datamart_data,
         )
     elif task['augmentation']['type'] == 'union':
         output_metadata = union(
-            pd.read_csv(io.BytesIO(data), error_bad_lines=False),
+            pd.read_csv(data, error_bad_lines=False),
             newdata,
             metadata,
             task['metadata'],
